unit MainForm_EMUZ80;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, FileUtil, LResources, Forms, Controls,
  Graphics, Dialogs, ExtCtrls, IniFiles, dateutils,

  {$IFDEF MSWINDOWS}
     MMSystem,
  {$ENDIF}

  MainForm_buttons, Emulator, MainForm_ram, MainForm_cpu, MainForm_trace,
  modul_cpu, modul_ram, modul_pio, wincpm;



  const
    max_Zeilen = 25;
    max_Spalten = 81;
    x_Offset = 3;
    y_Offset = 29;
    // Fuellzeichen = '.';
    Fuellzeichen = ' ';
    VG_Farbe = clBlack;
    HG_Farbe = clWhite;
    CH_Style = [];
    ZeiHoehe = 18;
    //  ZeiBreite = 10;
    ZeiBreite  = 11;
    LeftMargin = 40;
    TopMargin  = 24;

    CAP_Text = 'EMUZ80 - Z80 Emulator von Ronald Daleske, Version 1.05';


    {$IFDEF MSWINDOWS}
       Pfadtrenner = '\';
    {$ELSE}
       Pfadtrenner = '/';
    {$ENDIF}


{$I CBIOS_TXT.INC}

{$I BDOS_TXT.INC}

{$I CCP_TXT.INC}




  type

    TZeichenZeile = Array [1..max_Spalten] of Char;
    TZeichenMatrix = Array [1..max_Zeilen] of TZeichenZeile;

    TTermMode = (Normal_TM, IBM_ESC_TM, IBM_CSI_TM, IBM_CSI_TM_Par, IBM_CSI_TM_Par_Zeile,
                 PC1715_ESC_TM, PC1715_TM_Zeile, DCA_TM, DCA_TM_Zeile,
                 ADM31_ESC_TM, ADM31_ESC_TM_R, ADM31_ESC_TM_C, ADM31_ESC_ESC,
                 ADM31_ESC_ESC_ESC, ADM31_ESC_TM_G);

    TEmuMode = (Emu_VT100, Emu_VT52, Emu_PC1715, EMU_ANSI, EMU_ADM31);

    TMCursor = record
                 on_off : boolean;
                 Spalte, Zeile : Integer;
                 Zeichen : Char;
               end;




  { ***** Form_EMUZ80 ***** }

  { TForm_EMUZ80 }

  TForm_EMUZ80 = class(TForm)
    Image1: TImage;
    Image_Disk_LW_B: TImage;
    Image_Button_Einstellungen: TImage;
    Image_Button_Schliessen: TImage;
    Image_Disk_LW_A: TImage;
    Image_Graf_Mon2: TImage;
    Image_Graf_Mon1: TImage;
    Panel1: TPanel;
    Timer_Bild_Refresh: TTimer;
    Timer_Step: TTimer;
    Timer_CURSOR: TTimer;

    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure FormKeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
    procedure FormShow(Sender: TObject);
    procedure Image_Button_EinstellungenClick(Sender: TObject);
    procedure Image_Button_SchliessenClick(Sender: TObject);
    procedure Timer_Bild_RefreshTimer(Sender: TObject);
    procedure Timer_CURSORTimer(Sender: TObject);
    procedure Timer_StepTimer(Sender: TObject);


  private
    { private declarations }

    FIni:TMemIniFile;
    MCursor : TMCursor;
    ZeichenMatrix : TZeichenMatrix;
    FarbMatrix : Array [1..max_Spalten,1..max_Zeilen] of TColor;
    StyleMatrix : Array [1..max_Spalten,1..max_Zeilen] of TFontStyles;
    TermMode : TTermMode;
    EmuMode : TEmuMode;
    CSI_Parameter1, CSI_Parameter2, ESC_Zeile : Integer;
    CONOUT_Line : String;
    CONTROL_ACTIVE : Boolean;
    FCOLOR, BCOLOR : TColor;
    CSTYLE : TFontStyles;
    Emu_Mode : Integer;

    Stop_Bit_ESC_Sequenz : Boolean;

    warte_trace : Boolean;

    procedure RUN_DEL;
    procedure Matrix_in_GrafMon;
    procedure Lade_BIOS_in_RAM;


  public
    { public declarations }
    Z80 : TZ80;
    Main_Left, Main_Top,
    Mem_Left, Mem_Top,
    Button_Left, Button_Top,
    CPU_Left, CPU_Top,
    TRACE_Left, TRACE_Top,
    TRACE_Width, TRACE_Height,
    VAR_SECTOR, VAR_TRACK, VAR_DISKNO : Integer;
    Programm_File_Pfad, CPM_DIR_LW_A,
    CPM_DIR_LW_B : WideString;
    LOG_READ, CPM_AutoStart,
    EMU_Mode_Trace, EMU_Mode_Run,
    Form_RAM_aktiv, Form_CPU_aktiv,
    Form_Button_aktiv,
    Form_TRACE_aktiv : Boolean;

    procedure SET_Graf_Term;
    procedure Zeichen_an_CONOUT(Zeichen : Byte);
    procedure Reset_Form;
    procedure Read_IMAGE_Sektor(ISEC_LFD_NR : Byte;
                                var DMAAD : Word);
    procedure Write_IMAGE_Sektor(ISEC_LFD_NR : Byte;
                                 var DMAAD : Word);
    procedure Lade_BDOS_in_RAM(ADR_HI : Byte);
    procedure Lade_CCP_in_RAM(ADR_HI : Byte);

  end;



var
  Form_EMUZ80: TForm_EMUZ80;

implementation




{ ##### TForm_EMUZ80 ##### }

procedure TForm_EMUZ80.FormCreate(Sender: TObject);
begin
  Z80 := TZ80.Create;


  Programm_File_Pfad := ExtractFilePath(Application.ExeName);
  FIni := TMemIniFile.Create(Programm_File_Pfad +'EMUZ80.ini');
  Main_Left := FIni.ReadInteger('FORM', 'Main_Left', 100);
  Main_Top  := FIni.ReadInteger('FORM', 'Main_Top', 100);

  Button_Left  := FIni.ReadInteger('FORM', 'Button_Left', 100);
  Button_Top   := FIni.ReadInteger('FORM', 'Button_Top', 100);

  Mem_Left  := FIni.ReadInteger('FORM', 'Mem_Left', 100);
  Mem_Top   := FIni.ReadInteger('FORM', 'Mem_Top', 100);
  CPU_Left  := FIni.ReadInteger('FORM', 'CPU_Left', 100);
  CPU_Top   := FIni.ReadInteger('FORM', 'CPU_Top', 100);
  TRACE_Left:= FIni.ReadInteger('FORM', 'TRACE_Left', 100);
  TRACE_Top := FIni.ReadInteger('FORM', 'TRACE_Top', 100);
  TRACE_Width := FIni.ReadInteger('FORM', 'TRACE_Width', 100);
  TRACE_Height := FIni.ReadInteger('FORM', 'TRACE_Height', 100);
  CPM_AutoStart := FIni.ReadBool('FORM', 'CPM_AutoStart', true);
  Emu_Mode := FIni.ReadInteger('Emulation', 'Emu_Mode', 0);

  CPM_DIR_LW_A := FIni.ReadString('CPM', 'CPM_DIR_LW_A', 'CPM_LW_A');
  CPM_DIR_LW_B := FIni.ReadString('CPM', 'CPM_DIR_LW_B', 'CPM_LW_B');

  mod_pio := Tmod_pio.Create;



  FCOLOR:=clBlack;
  BCOLOR:=clWhite;
  CSTYLE:=[];


  Image_Graf_Mon1.Left:=40;
  Image_Graf_Mon1.Top:=40;
  Image_Graf_Mon1.Height:=509;
  Image_Graf_Mon1.Width:=910;
  Image_Graf_Mon1.Canvas.Brush.Color:=clWhite;
  Image_Graf_Mon1.Canvas.Pen.Color:=clWhite;
  Image_Graf_Mon1.Canvas.Pen.Width:=1;
  Image_Graf_Mon1.Canvas.Font.Name:='Courier New';
  Image_Graf_Mon1.Canvas.Font.Size:=12;
  Image_Graf_Mon1.Canvas.Font.Color:=FCOLOR;
  Image_Graf_Mon1.Canvas.Rectangle(0,0,Image_Graf_Mon1.Width,Image_Graf_Mon1.Height);


  Image_Graf_Mon2.Left:=40;
  Image_Graf_Mon2.Top:=40;
  Image_Graf_Mon2.Height:=509;
  Image_Graf_Mon2.Width:=910;
  Image_Graf_Mon2.Canvas.Brush.Color:=clWhite;
  Image_Graf_Mon2.Canvas.Pen.Color:=clWhite;
  Image_Graf_Mon2.Canvas.Pen.Width:=1;
  Image_Graf_Mon2.Canvas.Font.Name:='Courier New';
  Image_Graf_Mon2.Canvas.Font.Size:=12;
  Image_Graf_Mon2.Canvas.Font.Color:=FCOLOR;
  Image_Graf_Mon2.Canvas.Rectangle(0,0,Image_Graf_Mon1.Width,Image_Graf_Mon1.Height);


  SET_Graf_Term;


  case Emu_Mode of
    0 : EmuMode := Emu_VT100;
    1 : EmuMode := Emu_VT52;
    2 : EmuMode := Emu_PC1715;
    3 : EmuMode := EMU_ANSI;
    4 : EmuMode := EMU_ADM31;
  else
    EmuMode := Emu_VT100;
  end;

  Reset_Form;

  CONOUT_Line:='';
  CONTROL_ACTIVE:=false;

  EMU_Mode_Trace := false;
  EMU_Mode_Run := false;

  LOG_READ := false;

  Form_RAM_aktiv:=false;
  Form_CPU_aktiv:=false;
  Form_TRACE_aktiv:=false;
  Form_Button_aktiv:=false;

  Caption:=CAP_Text;

  Stop_Bit_ESC_Sequenz:=false;

  warte_trace:=false;

end; { procedure TForm_EMUZ80.FormCreate }






procedure TForm_EMUZ80.FormShow(Sender: TObject);
begin

  Form_EMUZ80.Left := Main_Left;
  Form_EMUZ80.Top := Main_Top;
  Form_RAM.Left := Mem_Left;
  Form_RAM.Top := Mem_Top;
  Form_cpu.Left := CPU_Left;
  Form_cpu.Top := CPU_Top;
  Form_trace.Left := TRACE_Left;
  Form_trace.Top := TRACE_Top;

  TForm_BUTTONS.Left:=Button_Left;
  TForm_BUTTONS.Top:=Button_Top;

  TForm_BUTTONS.CheckBox_CPM_AutoStart.Checked := CPM_AutoStart;



  Lade_BIOS_in_RAM;
  mod_cpu.Reg.SRS.PC:=$00;




  if CPM_AutoStart then begin

    EMU_Mode_Run:=true;
    EMU_Mode_Trace:=false;
    Z80.Mode_Protokoll:=false;

    TForm_BUTTONS.Button_2_ein;

  end; { if CPM_AutoStart then }


end; { procedure TForm_EMUZ80.FormShow }














procedure TForm_EMUZ80.Lade_BIOS_in_RAM;
var
  A_Zaehler : Integer;
begin

  for A_Zaehler:=0 to Laenge_CBIOS_Array do begin
    mod_ram.Write_Byte(A_Zaehler,CBIOS_ARR[A_Zaehler]);
  end; { for A_Zaehler:=0 to Laenge_CBIOS_Array do }

end; { procedure TForm_EMUZ80.Lade_BIOS_in_RAM }







procedure TForm_EMUZ80.Lade_BDOS_in_RAM(ADR_HI : Byte);
var
  A_Zaehler : Integer;
  Adresse : Word;
begin
  Adresse := ADR_HI shl 8;

  for A_Zaehler:=0 to Laenge_BDOS_Array do begin
    mod_ram.Write_Byte(Adresse+A_Zaehler,BDOS_ARR[A_Zaehler]);
  end; { for A_Zaehler:=0 to Laenge_BDOS_Array do }

end; { procedure TForm_EMUZ80.Lade_BDOS_in_RAM }









procedure TForm_EMUZ80.Lade_CCP_in_RAM(ADR_HI : Byte);
var
  A_Zaehler : Integer;
  Adresse : Word;
begin
  Adresse := ADR_HI shl 8;

  for A_Zaehler:=0 to Laenge_CCP_Array do begin
    mod_ram.Write_Byte(Adresse+A_Zaehler,CCP_ARR[A_Zaehler]);
  end; { for A_Zaehler:=0 to Laenge_BDOS_Array do }

end; { procedure TForm_EMUZ80.Lade_CCP_in_RAM }









procedure TForm_EMUZ80.Read_IMAGE_Sektor(ISEC_LFD_NR : Byte;
                                         var DMAAD : Word);
var
  ein_Byte : Byte;
  x1 : Integer;
  I_POS : LongInt;
  Zeile, Zeile2 : String;

begin


  if LOG_READ and (ISEC_LFD_NR=0) then
    TForm_BUTTONS.Memo_Protokoll.Lines.Add('READ - Sec='+IntToStr(VAR_SECTOR)+
                       ' Trk='+IntToStr(VAR_TRACK)+'  Dsk='+IntToStr(VAR_DISKNO));


  case VAR_DISKNO of
    0 : begin
          I_POS:=(ISEC_LFD_NR*32)+(((VAR_TRACK*wcpm_LW_A.DPB.SPT)+(VAR_SECTOR))*128);
          wcpm_LW_A.IMG_File.Position:=I_POS;
          if (I_POS>wcpm_LW_A.IMG_File.Size) then begin
            TForm_BUTTONS.Memo_Protokoll.Lines.Add('*** Fehler max Groesse Laufwerk A: ueberschritten='+IntToStr(I_POS));
            sleep(2000);
          end;
        end; { 0 }
    1 : begin
          I_POS:=(ISEC_LFD_NR*32)+(((VAR_TRACK*wcpm_LW_B.DPB.SPT)+(VAR_SECTOR))*128);
          wcpm_LW_B.IMG_File.Position:=I_POS;
          if (I_POS>wcpm_LW_B.IMG_File.Size) then begin
            TForm_BUTTONS.Memo_Protokoll.Lines.Add('*** Fehler max Groesse Laufwerk B: ueberschritten='+IntToStr(I_POS));
            sleep(2000);
          end;
        end; { 1 }
  else
    TForm_BUTTONS.Memo_Protokoll.Lines.Add('*** Fehler, Das Laufwerk: '+IntToStr(VAR_DISKNO)+': ist nicht installiert!');
  end; { case VAR_DISKNO of }


  Zeile:='';
  Zeile2:='';

  for x1:=0 to 31 do begin

    case VAR_DISKNO of
      0 : wcpm_LW_A.IMG_File.Read(ein_Byte,1);
      1 : wcpm_LW_B.IMG_File.Read(ein_Byte,1);
    else
      ein_Byte:=$E5;
    end; { case VAR_DISKNO of }

    mod_ram.Write_Byte(DMAAD, ein_Byte);
    INC(DMAAD);

    Zeile:=Zeile+IntToHex(ein_Byte,2)+'H,';

    if (ein_Byte>=$20) and (ein_Byte<$80) then
      Zeile2:=Zeile2+Char(ein_Byte)+'   '
    else
      Zeile2:=Zeile2+'    ';

  end; { for x1:=0 to 31 do }


  if LOG_READ then begin
    TForm_BUTTONS.Memo_Protokoll.Lines.Add(Zeile);
    TForm_BUTTONS.Memo_Protokoll.Lines.Add(Zeile2);
  end;


end; { procedure TForm_EMUZ80.Read_IMAGE_Sektor }








procedure TForm_EMUZ80.Write_IMAGE_Sektor(ISEC_LFD_NR : Byte;
                                          var DMAAD : Word);
var

  ein_Byte : Byte;
  x1 : Integer;
  I_POS : LongInt;
  Zeile, Zeile2 : String;

begin


  if LOG_READ and (ISEC_LFD_NR=0) then
    TForm_BUTTONS.Memo_Protokoll.Lines.Add('WRITE - Sec='+IntToStr(VAR_SECTOR)+
                       ' Trk='+IntToStr(VAR_TRACK)+'  Dsk='+IntToStr(VAR_DISKNO));

  case VAR_DISKNO of
    0 : begin
          I_POS:=(ISEC_LFD_NR*32)+(((VAR_TRACK*wcpm_LW_A.DPB.SPT)+(VAR_SECTOR))*128);
          wcpm_LW_A.IMG_File.Position:=I_POS;
          if (I_POS>wcpm_LW_A.IMG_File.Size) then begin
            TForm_BUTTONS.Memo_Protokoll.Lines.Add('*** Fehler max Groesse Laufwerk A: ueberschritten='+IntToStr(I_POS));
            sleep(2000);
          end;
        end; { 0 }
    1 : begin
          I_POS:=(ISEC_LFD_NR*32)+(((VAR_TRACK*wcpm_LW_B.DPB.SPT)+(VAR_SECTOR))*128);
          wcpm_LW_B.IMG_File.Position:=I_POS;
          if (I_POS>wcpm_LW_B.IMG_File.Size) then begin
            TForm_BUTTONS.Memo_Protokoll.Lines.Add('*** Fehler max Groesse Laufwerk B: ueberschritten='+IntToStr(I_POS));
            sleep(2000);
          end;
        end; { 1 }
  else
    TForm_BUTTONS.Memo_Protokoll.Lines.Add('*** Fehler, Das Laufwerk: '+IntToStr(VAR_DISKNO+
                                               StrToInt('A'))+': ist nicht installiert!');
  end; { case VAR_DISKNO of }



  Zeile:='';
  Zeile2:='';

  for x1:=0 to 31 do begin

    ein_Byte := mod_ram.Read_Byte(DMAAD);
    INC(DMAAD);

    case VAR_DISKNO of
      0 : wcpm_LW_A.IMG_File.Write(ein_Byte,1);
      1 : wcpm_LW_B.IMG_File.Write(ein_Byte,1);
    else
      TForm_BUTTONS.Memo_Protokoll.Lines.Add('*** Fehler, Das Laufwerk: '+IntToStr(VAR_DISKNO+
                                             StrToInt('A'))+': ist nicht installiert!');
    end; { case VAR_DISKNO of }



    Zeile:=Zeile+IntToHex(ein_Byte,2)+'H,';

    if (ein_Byte>=$20) and (ein_Byte<$80) then
      Zeile2:=Zeile2+Char(ein_Byte)+'   '
    else
      Zeile2:=Zeile2+'    ';


  end; { for x1:=0 to 31 do }



  if LOG_READ then begin
    TForm_BUTTONS.Memo_Protokoll.Lines.Add(Zeile);
    TForm_BUTTONS.Memo_Protokoll.Lines.Add(Zeile2);
  end;


end; { procedure TForm_EMUZ80.Write_IMAGE_Sektor }






{

  TIMER TIMER TIMER TIMER TIMER TIMER TIMER TIMER TIMER TIMER

}








procedure TForm_EMUZ80.Timer_CURSORTimer(Sender: TObject);
begin
  Timer_CURSOR.Enabled:=false;

  // Ist der Kursor eingeschaltet (MCursor.on_off:=true;  und "_" wird angezeigt),
  // dann befindet sich auch in MCursor.Zeichen das Zeichen
  if MCursor.on_off then begin
    ZeichenMatrix[MCursor.Zeile,MCursor.Spalte]:=MCursor.Zeichen;
    MCursor.on_off:=false;
  end
  else begin
    MCursor.Zeichen:=ZeichenMatrix[MCursor.Zeile,MCursor.Spalte];
    ZeichenMatrix[MCursor.Zeile,MCursor.Spalte]:='_';
    MCursor.on_off:=true;
  end;

  Timer_CURSOR.Enabled:=true;

end; { procedure TForm_EMUZ80.Timer_CURSORTimer }







procedure TForm_EMUZ80.Timer_StepTimer(Sender: TObject);
var
  MR_Zaehler : Integer;
begin

  Timer_Step.Enabled:=false;


  if EMU_Mode_Trace then begin

    Z80.einen_Befehl_abarbeiten;

    if Z80.Mode_Protokoll then begin

      Form_RAM.Inhalt_anzeigen;
      Form_cpu.Register_anzeigen;

    end; { if Z80.Mode_Protokoll then }

  end;  { if EMU_Trace_on then }



  MR_Zaehler := 30000;
  while EMU_Mode_Run and (MR_Zaehler>0) do begin

    Z80.einen_Befehl_abarbeiten;

    DEC(MR_Zaehler);

  end;  { if Mode_Run then }

  Timer_Step.Enabled:=true;

end; { procedure TForm_EMUZ80.Timer_StepTimer }








procedure TForm_EMUZ80.Timer_Bild_RefreshTimer(Sender: TObject);
begin

  Timer_Bild_Refresh.Enabled:=false;

  Matrix_in_GrafMon;

  Timer_Bild_Refresh.Enabled:=true;


end; { procedure TForm_EMUZ80.Timer_Bild_RefreshTimer }







{ Console Console Console Console  }







procedure TForm_EMUZ80.Matrix_in_GrafMon;
var
  z1, s1, x1, y1 : Integer;
  MatrixZeichen : Char;
begin
  x1:=LeftMargin;
  y1:=TopMargin;
  for z1:=1 to max_Zeilen do begin
    for s1:=1 to max_Spalten do begin
      MatrixZeichen := ZeichenMatrix[z1,s1];
      if Image_Graf_Mon1.Visible then begin
        Image_Graf_Mon2.Canvas.Font.Color:=FarbMatrix[z1,s1];
        Image_Graf_Mon2.Canvas.Font.Style:=StyleMatrix[z1,s1];
        x1:=LeftMargin+(ZeiBreite*(s1-1));
        y1:=TopMargin+(ZeiHoehe*(z1-1));
        Image_Graf_Mon2.Canvas.TextOut(x1,y1,MatrixZeichen);
      end
      else begin
        Image_Graf_Mon1.Canvas.Font.Color:=FarbMatrix[z1,s1];
        Image_Graf_Mon1.Canvas.Font.Style:=StyleMatrix[z1,s1];
        x1:=LeftMargin+(ZeiBreite*(s1-1));
        y1:=TopMargin+(ZeiHoehe*(z1-1));
        Image_Graf_Mon1.Canvas.TextOut(x1,y1,MatrixZeichen);
      end;

    end;

  end;
  if Image_Graf_Mon1.Visible then begin
    Image_Graf_Mon1.Visible:=false;
    Image_Graf_Mon2.Visible:=true;
  end
  else begin
    Image_Graf_Mon2.Visible:=false;
    Image_Graf_Mon1.Visible:=true;
  end;
end; { procedure TForm_EMUZ80.Matrix_in_GrafMon }








procedure TForm_EMUZ80.SET_Graf_Term;
begin

  Image_Graf_Mon1.Visible:=true;
  Image_Graf_Mon2.Visible:=false;

end; { procedure TForm_EMUZ80.SET_Graf_Term }









procedure TForm_EMUZ80.Reset_Form;
var
  z1, s1 : Integer;
begin

  for s1:=1 to max_Spalten do begin
    for z1:=1 to max_Zeilen do begin
      ZeichenMatrix[z1,s1]:=Fuellzeichen;
    end;
  end;


  for s1:=1 to max_Spalten do begin
    for z1:=1 to max_Zeilen do begin
      FarbMatrix[z1,s1]:=FCOLOR;
    end;
  end;


  for s1:=1 to max_Spalten do begin
    for z1:=1 to max_Zeilen do begin
      StyleMatrix[z1,s1]:=CSTYLE;
    end;
  end;


  MCursor.Spalte:=1;
  MCursor.Zeile:=1;
  MCursor.on_off:=false;

  CSI_Parameter1:=0;
  CSI_Parameter2:=0;
  ESC_Zeile:=0;
  TermMode := Normal_TM;

  Timer_CURSOR.Enabled:=true;

end; { procedure TForm_EMUZ80.Reset_Form }










procedure TForm_EMUZ80.RUN_DEL;
begin
  Timer_CURSOR.Enabled:=false;

  if MCursor.on_off then begin
    ZeichenMatrix[MCursor.Zeile,MCursor.Spalte]:=MCursor.Zeichen;
    MCursor.on_off:=false;
  end; { if MCursor.on_off then }

  if MCursor.Spalte>1 then begin
    dec(MCursor.Spalte);
    ZeichenMatrix[MCursor.Zeile,MCursor.Spalte]:=Fuellzeichen;
    FarbMatrix[MCursor.Zeile,MCursor.Spalte]:=VG_Farbe;
    StyleMatrix[MCursor.Zeile,MCursor.Spalte]:=CH_Style;
  end;

  Timer_CURSOR.Enabled:=true;
end; { procedure TForm_EMUZ80.RUN_DEL }









procedure TForm_EMUZ80.Zeichen_an_CONOUT(Zeichen : Byte);




  procedure RUN_BELL;
  begin

    {$IFDEF MSWINDOWS}
       sndPlaySound(PChar('bell.wav'), SND_NODEFAULT Or SND_ASYNC);
    {$ENDIF}

  end; { procedure RUN_BELL }




  procedure scroll_down;
  var
    z1, s1 : Integer;
  begin
    for z1:=1 to max_Zeilen-1 do begin
      ZeichenMatrix[z1]:=ZeichenMatrix[z1+1];
      FarbMatrix[z1]:=FarbMatrix[z1+1];
      StyleMatrix[z1]:=StyleMatrix[z1+1];
    end;
    for s1:=1 to max_Spalten do begin
      ZeichenMatrix[max_Zeilen,s1]:=Fuellzeichen;
      FarbMatrix[max_Zeilen,s1]:=VG_Farbe;
      StyleMatrix[max_Zeilen,s1]:=CH_Style;
    end;
    MCursor.Spalte:=1;
    MCursor.Zeile:=max_Zeilen;
  end; { procedure scroll_down }




  procedure erase_screen;
  var
    z1, s1 : Integer;
  begin

    for s1:=1 to max_Spalten do begin
      for z1:=1 to max_Zeilen do begin
        ZeichenMatrix[z1,s1]:=Fuellzeichen;
      end;
    end;


    for s1:=1 to max_Spalten do begin
      for z1:=1 to max_Zeilen do begin
        FarbMatrix[z1,s1]:=FCOLOR;
      end;
    end;


    for s1:=1 to max_Spalten do begin
      for z1:=1 to max_Zeilen do begin
        StyleMatrix[z1,s1]:=CSTYLE;
      end;
    end;
  end; { procedure erase_screen }



  procedure Zeichen_in_Matrix(MZeichen : Byte);
  begin
    ZeichenMatrix[MCursor.Zeile,MCursor.Spalte]:=Chr(MZeichen);
    FarbMatrix[MCursor.Zeile,MCursor.Spalte]:=FCOLOR;
    StyleMatrix[MCursor.Zeile,MCursor.Spalte]:=CSTYLE;

    Inc(MCursor.Spalte);
    if MCursor.Spalte>max_Spalten then begin
      MCursor.Spalte:=1;
      Inc(MCursor.Zeile);
      if MCursor.Zeile>max_Zeilen then begin
        scroll_down;
      end;
    end;

  end; { procedure Zeichen_in_Matrix }



  procedure SonderZeichen_in_Matrix;
  var
    SZeile : String;
  begin
    SZeile:='['+IntToHex(Zeichen,2)+']';
    TForm_BUTTONS.Memo_Protokoll.Lines.Add('CODE : '+SZeile+' nicht aufgeloest');

  end; { procedure SonderZeichen_in_Matrix }



  procedure ESC_SonderZeichen_in_Matrix(SonderZeichen : String);
  var
    SZeile : String;
    SZaehler, SondZeichen : Integer;
    CZeichen : Char;
  begin
    SZeile:='[ESC]';
    for SZaehler:=1 to Length(SonderZeichen) do begin
      CZeichen := SonderZeichen[SZaehler];
      SondZeichen := ord(CZeichen);
      SZeile:=SZeile+'['+IntToHex(SondZeichen,2)+']';
    end;

    SZeile:=SZeile+'['+IntToHex(Zeichen,2)+']';
    TForm_BUTTONS.Memo_Protokoll.Lines.Add('CODE : '+SZeile+' nicht aufgeloest');

  end; { procedure ESC_SonderZeichen_in_Matrix }






  procedure CURSOR_UP;
  begin
    if MCursor.Zeile>1 then
      Dec(MCursor.Zeile);

    MCursor.on_off:=true;
    TermMode:=Normal_TM;
  end; { procedure CURSOR_UP }


  procedure CURSOR_DOWN;
  begin
    if MCursor.Zeile<max_Zeilen then
      Inc(MCursor.Zeile);

    MCursor.on_off:=true;
    TermMode:=Normal_TM;
  end; { procedure CURSOR_DOWN }


  procedure CURSOR_RIGHT;
  begin
    if MCursor.Spalte<max_Spalten then
      Inc(MCursor.Spalte);

    MCursor.on_off:=true;
    TermMode:=Normal_TM;
  end; { procedure CURSOR_RIGHT }


  procedure CURSOR_LEFT;
  begin
    if MCursor.Spalte>1 then
      Dec(MCursor.Spalte);

    MCursor.on_off:=true;
    TermMode:=Normal_TM;
  end; { procedure CURSOR_LEFT }


  procedure CURSOR_HOME;
  begin
    MCursor.Spalte:=1;
    MCursor.Zeile:=1;
    MCursor.on_off:=true;
    TermMode:=Normal_TM;
  end; { procedure CURSOR_HOME }



  procedure DELETE_EOL;
  var
    s1 : Integer;
  begin
    for s1:=MCursor.Spalte to max_Spalten do begin
      ZeichenMatrix[MCursor.Zeile,s1]:=Fuellzeichen;
      FarbMatrix[MCursor.Zeile,s1]:=VG_Farbe;
      StyleMatrix[MCursor.Zeile,s1]:=CH_Style;
    end;

    MCursor.on_off:=true;
    TermMode:=Normal_TM;
  end; { procedure DELETE_EOL }




  procedure DELETE_CHARACTER;
  var
    s1 : Integer;
  begin
    for s1:=MCursor.Spalte to max_Spalten-1 do begin
      ZeichenMatrix[MCursor.Zeile,s1]:=ZeichenMatrix[MCursor.Zeile,s1+1];
      FarbMatrix[MCursor.Zeile,s1]:=FarbMatrix[MCursor.Zeile,s1+1];
      StyleMatrix[MCursor.Zeile,s1]:=StyleMatrix[MCursor.Zeile,s1+1];
    end;

    MCursor.on_off:=true;
    TermMode:=Normal_TM;
  end; { procedure DELETE_CHARACTER }



  procedure DELETE_BOL;
  var
    s1 : Integer;
  begin
    for s1:=1 to MCursor.Spalte-1 do begin
      ZeichenMatrix[MCursor.Zeile,s1]:=Fuellzeichen;
      FarbMatrix[MCursor.Zeile,s1]:=VG_Farbe;
      StyleMatrix[MCursor.Zeile,s1]:=CH_Style;
    end;

    MCursor.on_off:=true;
    TermMode:=Normal_TM;
  end; { procedure DELETE_BOL }



  procedure DELETE_LINE;
  var
    s1 : Integer;
  begin
    for s1:=1 to max_Spalten do begin
      ZeichenMatrix[MCursor.Zeile,s1]:=Fuellzeichen;
      FarbMatrix[MCursor.Zeile,s1]:=VG_Farbe;
      StyleMatrix[MCursor.Zeile,s1]:=CH_Style;
    end;

    MCursor.Spalte:=1;
    MCursor.on_off:=true;
    TermMode:=Normal_TM;
  end; { procedure DELETE_LINE }




  procedure DELETE_LINE_SCR_ROLL;
  var
    z1, s1 : Integer;
  begin
    for z1:=MCursor.Zeile to max_Zeilen-1 do begin
      ZeichenMatrix[z1]:=ZeichenMatrix[z1+1];
      FarbMatrix[z1]:=FarbMatrix[z1+1];
      StyleMatrix[z1]:=StyleMatrix[z1+1];
    end;
    for s1:=1 to max_Spalten do begin
      ZeichenMatrix[max_Zeilen,s1]:=Fuellzeichen;
      FarbMatrix[max_Zeilen,s1]:=VG_Farbe;
      StyleMatrix[max_Zeilen,s1]:=CH_Style;
    end;
    MCursor.Spalte:=1;
    MCursor.on_off:=true;
    TermMode:=Normal_TM;
  end; { procedure DELETE_LINE_SCR_ROLL }



  procedure DELETE_EOS;
  var
    s1, z1 : Integer;
  begin
    DELETE_EOL;

    if MCursor.Zeile<max_Zeilen then begin
      for z1:=MCursor.Zeile+1 to max_Zeilen do begin
         for s1:=1 to max_Spalten do begin
           ZeichenMatrix[z1,s1]:=Fuellzeichen;
           FarbMatrix[z1,s1]:=VG_Farbe;
           StyleMatrix[z1,s1]:=CH_Style;
         end;
      end;
    end;

    MCursor.on_off:=true;
    TermMode:=Normal_TM;
  end; { procedure DELETE_EOS }




  procedure DELETE_BOS;
  var
    s1, z1 : Integer;
  begin
    DELETE_BOL;

    if MCursor.Zeile>1 then begin
      for z1:=1 to MCursor.Zeile-1 do begin
         for s1:=1 to max_Spalten do begin
           ZeichenMatrix[z1,s1]:=Fuellzeichen;
           FarbMatrix[z1,s1]:=VG_Farbe;
           StyleMatrix[z1,s1]:=CH_Style;
         end;
      end;
    end;

    MCursor.on_off:=true;
    TermMode:=Normal_TM;
  end; { procedure DELETE_BOS }



  procedure LINE_FEED;
  begin
    Inc(MCursor.Zeile);

    if MCursor.Zeile>max_Zeilen then begin
      scroll_down;
    end;

    TermMode:=Normal_TM;
  end; { procedure LINE_FEED }




  procedure INSERT_CHARACTER;
  var
    s1 : Integer;
  begin
    for s1:=max_Spalten downto MCursor.Spalte+1 do begin
      ZeichenMatrix[MCursor.Zeile,s1]:=ZeichenMatrix[MCursor.Zeile,s1-1];
      FarbMatrix[MCursor.Zeile,s1]:=FarbMatrix[MCursor.Zeile,s1-1];
      StyleMatrix[MCursor.Zeile,s1]:=StyleMatrix[MCursor.Zeile,s1-1];
    end;
    ZeichenMatrix[MCursor.Zeile,MCursor.Spalte]:=Fuellzeichen;
    FarbMatrix[MCursor.Zeile,MCursor.Spalte]:=VG_Farbe;
    StyleMatrix[MCursor.Zeile,MCursor.Spalte]:=CH_Style;
    TermMode:=Normal_TM;
  end; { procedure INSERT_CHARACTER }






  procedure INSERT_LINE_SCR_ROLL;
  var
    z1, s1 : Integer;
  begin
    for z1:=max_Zeilen downto MCursor.Zeile+1 do begin
      ZeichenMatrix[z1]:=ZeichenMatrix[z1-1];
      FarbMatrix[z1]:=FarbMatrix[z1-1];
      StyleMatrix[z1]:=StyleMatrix[z1-1];
    end;
    for s1:=1 to max_Spalten do begin
      ZeichenMatrix[MCursor.Zeile,s1]:=Fuellzeichen;
      FarbMatrix[MCursor.Zeile,s1]:=VG_Farbe;
      StyleMatrix[MCursor.Zeile,s1]:=CH_Style;
    end;
    MCursor.on_off:=true;
    TermMode:=Normal_TM;
  end; { procedure INSERT_LINE_SCR_ROLL }






  procedure Normal_Terminal_Mode;
  begin

    case Zeichen of
      $01 : CURSOR_HOME;   // auf Pos 1,1
      $04 : CURSOR_RIGHT;  // rechts
      $05 : CURSOR_UP;     // oben
      $07 : RUN_BELL;      // BELL - Glocke
      $09 : MCursor.Spalte:=8 * ((MCursor.Spalte div 8) +1);  // TAB
      $0A : LINE_FEED; // LF
      $0C : Reset_Form; // CLS - Clear Screen
      $0D : MCursor.Spalte:=1;  // CR
      $13 : CURSOR_LEFT;  // links
      $16 : DELETE_EOL;   // ab Cursorposition bis Zeilenende loeschen
      $18 : CURSOR_DOWN; // unten
      $1B : begin
              case EmuMode of
                Emu_PC1715 : TermMode:= PC1715_ESC_TM;
                Emu_VT100  : TermMode:=IBM_ESC_TM;
                Emu_VT52   : TermMode:=IBM_ESC_TM;
                EMU_ANSI   : TermMode:=IBM_ESC_TM;
                EMU_ADM31  : TermMode:=ADM31_ESC_TM;
              else
                TermMode:= PC1715_ESC_TM;
              end;
            end; { $1B [ESC] }

      $20..$7E : Zeichen_in_Matrix(Zeichen);

      $7F : RUN_DEL;
    else
      SonderZeichen_in_Matrix;
    end;

    if MCursor.on_off then begin
      MCursor.Zeichen:=ZeichenMatrix[MCursor.Zeile,MCursor.Spalte];
      ZeichenMatrix[MCursor.Zeile,MCursor.Spalte]:='_';
    end; { if MCursor.on_off then }

  end; { procedure Normal_Terminal_Mode }






  procedure SET_CURSOR_POSITION(Zeile, Spalte : Integer); // bezogen auf 0,0 !!!
  begin
    MCursor.Zeile:=Zeile;   //+1
    MCursor.Spalte:=Spalte;  //+1

    if (MCursor.Zeile>max_Zeilen) then
      MCursor.Zeile:=max_Zeilen;
                    {
    if (MCursor.Zeile<1) then
      MCursor.Zeile:=1;
                   }
    if (MCursor.Spalte>max_Spalten) then
      MCursor.Spalte:=max_Spalten;
                     {
    if (MCursor.Spalte<1) then
      MCursor.Spalte:=1;
                      }
    TermMode:=Normal_TM;

    // !!!!!!!!!!!!!!!!!!!!!!!!
    // MCursor.on_off:=true;


  end; { procedure SET_CURSOR_POSITION }



  // **********************************
  // ADM31_ESC_Terminal_Mode
  // **********************************


  procedure ADM31_ESC_Terminal_Mode;
  begin
    if Zeichen<>$1B then begin

      case Zeichen of

        $3D : begin
                TermMode:=ADM31_ESC_TM_R; // ESC = R C
              end; { $3D }
        $1B : begin
                TermMode:=ADM31_ESC_ESC; // ESC ESC
              end; { $3D }
        $54 : begin
                DELETE_EOL; // ESC T = clear to end of line
                TermMode:=Normal_TM;
              end; { $54 }
        $74 : begin
                DELETE_EOL; // ESC t = clear to end of line
                TermMode:=Normal_TM;
              end; { $74 }
        $59 : begin
                DELETE_EOS; // ESC Y = clear to end of screen
                TermMode:=Normal_TM;
              end; { $59 }
        $79 : begin
                DELETE_EOS; // ESC y = clear to end of screen
                TermMode:=Normal_TM;
              end; { $79 }
        $3A : begin
                Reset_Form; // ESC : = home & clear screen
                TermMode:=Normal_TM;
              end; { $3A }
        $2A : begin
                Reset_Form; // ESC * = home & clear screen
                TermMode:=Normal_TM;
              end; { $2A }
        $47 : begin
                TermMode:=ADM31_ESC_TM_G; // ESC G ...
              end; { $47 }
        $28 : begin
                // ESC ( = Half intensity off
                // derzeit nicht implementiert
              end; { $28 }
        $29 : begin
                // ESC ) = Half intensity on
                // derzeit nicht implementiert
              end; { $29 }
        $45 : begin
                LINE_FEED; // ESC E = Insert line
                TermMode:=Normal_TM;
              end; { $45 }
        $51 : begin
                INSERT_CHARACTER; // ESC Q = Insert Character
                TermMode:=Normal_TM;
              end; { $51 }
        $52 : begin
                DELETE_LINE; // ESC R = Delete Line
                TermMode:=Normal_TM;
              end; { $52 }
        $57 : begin
                DELETE_CHARACTER; // ESC W = Delete Character
                TermMode:=Normal_TM;
              end; { $57 }

      else
        ESC_SonderZeichen_in_Matrix(Char(Zeichen));
        TermMode:=Normal_TM;
        Normal_Terminal_Mode;
      end; { case Zeichen of }
    end
    else begin
      TForm_BUTTONS.Memo_Protokoll.Lines.Add('ESC+ESC empfangen');
    end;
  end; { procedure ADM31_ESC_Terminal_Mode }





  procedure ADM31_ESC_Par_Zeile_Terminal_Mode;
  begin
    if Zeichen>=$20 then begin
      ESC_Zeile:=Zeichen-$20;
      TermMode:=ADM31_ESC_TM_C;
    end
    else begin
      ESC_SonderZeichen_in_Matrix('=');
      TermMode:=Normal_TM;
    end;
  end; { procedure ADM31_ESC_Par_Zeile_Terminal_Mode }



  procedure ADM31_ESC_Par_Spalte_Terminal_Mode;
  begin
    if Zeichen>=$20 then begin
      SET_CURSOR_POSITION(ESC_Zeile, Zeichen-$20);
    end
    else begin
      ESC_SonderZeichen_in_Matrix('=R');
    end;
    TermMode:=Normal_TM;
  end; { procedure ADM31_ESC_Par_Spalte_Terminal_Mode }




  procedure ADM31_ESC_ESC_Terminal_Mode;
  begin
    case Zeichen of
      $1B : begin
              TermMode:=ADM31_ESC_ESC_ESC;
            end; { $1B }
      $3D : begin
              TermMode:=ADM31_ESC_TM_R; // ESC = R C
            end; { $3D }
      $29 : begin
              // Zeichen - nicht implementiert
            end; { $29 }
      $54 : begin
              // Zeichen - nicht implementiert
            end; { $54 }
    else
      ESC_SonderZeichen_in_Matrix(char($1B));
      TermMode:=Normal_TM;
    end;

  end; { procedure ADM31_ESC_ESC_Terminal_Mode }




  procedure ADM31_ESC_ESC_ESC_Terminal_Mode;
  begin
    if Zeichen=$1B then begin
      // color = Zeichen - nicht implementiert
      TermMode:=Normal_TM;
    end
    else begin
      ESC_SonderZeichen_in_Matrix(char($1B)+char($1B));
      TermMode:=Normal_TM;
    end;
  end; { procedure ADM31_ESC_ESC_ESC_Terminal_Mode }





  procedure ADM31_ESC_G_Terminal_Mode;
  begin
     case Zeichen of
       $00 : begin
               // ESC G 0 = Rev. video and blinking off, restore normal 80 col char set
               // nicht implementiert
             end; { $00 }
       $01 : begin
               // ESC G 1 = select 80 col alternate character set
               // nicht implementiert
             end; { $01 }
       $02 : begin
               // ESC G 2 = Blinking on
               // nicht implementiert
             end; { $02 }
       $04 : begin
               // ESC G 4 = Reverse video on
               // nicht implementiert
             end; { $04 }
      else
        ESC_SonderZeichen_in_Matrix(Char(Zeichen));
      end; { case Zeichen of }
     TermMode:=Normal_TM;
  end; { procedure ADM31_ESC_Terminal_Mode }




  // **********************************
  // IBM VT100 Terminal_Mode
  // **********************************




  procedure IBM_ESC_Terminal_Mode;
  begin
    case Zeichen of
      $3C : begin
              EmuMode:=Emu_VT100;  // VT52 = ESC < - ENTER_ANSI_MODE
            end; { $3C }
      $41 : begin
              EmuMode:=Emu_VT52;
              CURSOR_UP;  // VT52 = ESC A - CURSOR_UP
            end; { $41 }
      $42 : begin
              EmuMode:=Emu_VT52;
              CURSOR_DOWN; // VT52 = ESC B - CURSOR_DOWN
            end; { $42 }
      $43 : begin
              EmuMode:=Emu_VT52;
              CURSOR_RIGHT; // VT52 = ESC C - CURSOR_RIGHT
            end; { $43 }
      $44 : begin
              EmuMode:=Emu_VT52;
              CURSOR_LEFT; // VT52 = ESC D - CURSOR_LEFT
            end; { $44 }
      $48 : begin
              EmuMode:=Emu_VT52;
              CURSOR_HOME; // VT52 = ESC H - CURSOR_HOME
            end; { $48 }
      $4A : begin
              EmuMode:=Emu_VT52;
              DELETE_EOS; // VT52 = ESC J - DELETE_EOS
            end; { $4A }
      $4B : begin
              EmuMode:=Emu_VT52;
              DELETE_EOL; // VT52 = ESC K - DELETE_EOL
            end; { $4B }
      $59 : begin
              EmuMode:=Emu_VT52;
              TermMode:=DCA_TM; // VT52 = ESC Y Zeile Spalte - DIRECT_CURSOR_ADRESS
            end; { $59 }
      $5B : begin
              TermMode:=IBM_CSI_TM; // Control Sequence Introducer (CSI) = ESC [
              EmuMode:=Emu_VT100;
            end; { $5B }
    else
      ESC_SonderZeichen_in_Matrix(Char(Zeichen));
      TermMode:=Normal_TM;
    end; { case Zeichen of }
   end; { procedure ESC_Terminal_Mode }



  procedure IBM_CSI_Terminal_Mode;
  begin
    case Zeichen of
      $30..$39 : begin
                  CSI_Parameter1:=Zeichen-$30;
                  TermMode:=IBM_CSI_TM_Par;
                end;

      $48 : begin // Character Attributes = ESC [ H
              // set cursor home
              CURSOR_HOME;
              CSI_Parameter1:=0;
              TermMode:=Normal_TM;
            end; { $48 }

      $4A : begin // ESC [ J
              // erase to end of screen (inclusive)
              DELETE_EOS;
              CSI_Parameter1:=0;
              TermMode:=Normal_TM;
            end; { $4A }

      $4B : begin // ESC [ K
              // erase to end of line (inclusive)
              DELETE_EOL;
              CSI_Parameter1:=0;
              TermMode:=Normal_TM;
            end; { $4B }

      $4D : begin // ESC [ M
              // cursor up - at top of region, scroll down
              scroll_down;
              CURSOR_HOME;
              CSI_Parameter1:=0;
              TermMode:=Normal_TM;
            end; { $4D }

      $6D : begin // Character Attributes = ESC [ m
              // turn off attributes - normal video
              FCOLOR:=clBlack;
              BCOLOR:=clWhite;
              CSTYLE:=[];

              CSI_Parameter1:=0;
              TermMode:=Normal_TM;
            end; { $6D }

      $66 : begin // Character Attributes = ESC [ f
              // set cursor home
              CURSOR_HOME;
              CSI_Parameter1:=0;
              TermMode:=Normal_TM;
            end; { $66 }


    else
      ESC_SonderZeichen_in_Matrix('['+char(Zeichen));
      TermMode:=Normal_TM;
    end;
  end; { procedure IBM_CSI_Terminal_Mode }



  procedure IBM_CSI_Par_Terminal_Mode;
  var
    x_counter : Integer;
  begin
    case Zeichen of
      $30..$39 : begin
                  CSI_Parameter1:=(CSI_Parameter1*10)+Zeichen-$30;
                  TermMode:=IBM_CSI_TM_Par;
                end;

      $3B : begin // ";" -- MOVE CURSOR = ESC [ Pn1 ; PN2 H oder ESC [ Pn1 ; PN2 f
              TermMode:=IBM_CSI_TM_Par_Zeile;
            end; { $3B }

      $41 : begin // CUU -- Cursor Up = ESC [ Pn A
              if MCursor.Zeile>CSI_Parameter1 then
                MCursor.Zeile:=MCursor.Zeile-CSI_Parameter1
              else
                MCursor.Zeile:=1;

              TermMode:=Normal_TM;
              CSI_Parameter1:=0;
            end; { $41 }
      $42 : begin // CUD -- Cursor Down = ESC [ Pn B
              if (MCursor.Zeile+CSI_Parameter1)<max_Zeilen then
                MCursor.Zeile:=MCursor.Zeile+CSI_Parameter1
              else
                MCursor.Zeile:=max_Zeilen;

              CSI_Parameter1:=0;
              TermMode:=Normal_TM;
            end; { $42 }
      $43 : begin // CUF -- Cursor Forward  = ESC [ Pn C
              if (MCursor.Spalte+CSI_Parameter1)<max_Spalten then
                MCursor.Spalte:=MCursor.Spalte+CSI_Parameter1
              else
                MCursor.Spalte:=max_Spalten;

              CSI_Parameter1:=0;
              TermMode:=Normal_TM;
            end; { $43 }
      $44 : begin // CUB -- Cursor Backward = ESC [ Pn D
              if MCursor.Spalte>CSI_Parameter1 then
                MCursor.Spalte:=MCursor.Spalte-CSI_Parameter1
              else
                MCursor.Spalte:=1;

              CSI_Parameter1:=0;
              TermMode:=Normal_TM;
            end; { $44 }
      $4B : begin // ERASE LINE = ESC [ Pn K
              case CSI_Parameter1 of
                0 : DELETE_EOL;
                1 : DELETE_BOL;
                2 : DELETE_LINE;
              else
                ESC_SonderZeichen_in_Matrix('['+IntToStr(CSI_Parameter1)+'K');
              end;

              CSI_Parameter1:=0;
              TermMode:=Normal_TM;
            end; { $4B }

      $4A : begin // ERASE LINE = ESC [ Pn J
              case CSI_Parameter1 of
                // Löschen vom Bildanfang und von links bis zum Kursor
                0 : DELETE_EOS;   // erase to end of screen (inclusive)
                1 : DELETE_BOS;   // erase to beginning of screen (inclusive)
                2 : erase_screen; // erase entire screen (cursor doesn't move)
              else
                ESC_SonderZeichen_in_Matrix('['+IntToStr(CSI_Parameter1));
              end;

              CSI_Parameter1:=0;
              TermMode:=Normal_TM;
            end; { $4A }


      $4C : begin // ERASE LINE = ESC [ Pn L
              for x_counter:=0 to CSI_Parameter1-1 do
                INSERT_LINE_SCR_ROLL;

              CSI_Parameter1:=0;
              TermMode:=Normal_TM;
            end; { $4C }


      $4D : begin // ERASE LINE = ESC [ Pn M
              for x_counter:=0 to CSI_Parameter1-1 do
                DELETE_LINE_SCR_ROLL;

              CSI_Parameter1:=0;
              TermMode:=Normal_TM;
            end; { $4D }


      $6D : begin // Character Attributes = ESC [ Pn m
              case CSI_Parameter1 of

                0 : begin
                      CSTYLE:=[];            // turn off attributes - normal video
                      FCOLOR:=clBlack;
                      BCOLOR:=clWhite;
                    end;

                1 : begin
                      CSTYLE:=[fsBold];      // highlight
                      FCOLOR:=clBlack;
                      BCOLOR:=clWhite;
                    end;

                4 : begin
                      CSTYLE:=[fsUnderline]; // turn on underline mode
                      FCOLOR:=clBlack;
                      BCOLOR:=clWhite;
                    end;

                5 : begin
                      CSTYLE:=[fsItalic];    // blink
                      FCOLOR:=clBlack;
                      BCOLOR:=clWhite;
                    end;

                7 : begin
                      CSTYLE:=[fsBold];      // turn on inverse video mode
                      FCOLOR:=clBlue;
                      BCOLOR:=clWhite;
                    end;

              else
                ESC_SonderZeichen_in_Matrix('['+IntToStr(CSI_Parameter1));
              end;
              CSI_Parameter1:=0;
              TermMode:=Normal_TM;
            end; { $6D }
    else
      ESC_SonderZeichen_in_Matrix('['+IntToStr(CSI_Parameter1));
      CSI_Parameter1:=0;
      TermMode:=Normal_TM;
    end;

  end; { procedure IBM_CSI_Par_Terminal_Mode }






  procedure IBM_CSI_Par_Zeile_Terminal_Mode;
  begin
    case Zeichen of

      $30..$39 : begin
                   CSI_Parameter2:=(CSI_Parameter2*10)+Zeichen-$30;
                   TermMode:=IBM_CSI_TM_Par_Zeile;

                 end; { $30..$39 }

      $48 : begin // -- MOVE CURSOR = ESC [ Pn1 ; PN2 H
              SET_CURSOR_POSITION(CSI_Parameter1, CSI_Parameter2);
              CSI_Parameter1:=0;
              CSI_Parameter2:=0;
              TermMode:=Normal_TM;
            end; { $48 }

      $66 : begin // -- MOVE CURSOR = ESC [ Pn1 ; PN2 f
              SET_CURSOR_POSITION(CSI_Parameter1, CSI_Parameter2);
              CSI_Parameter1:=0;
              CSI_Parameter2:=0;
              TermMode:=Normal_TM;
            end; { $66 }

      $6D : begin // -- Character Attributes = ESC [ Pn1 ; PN2 m
              case CSI_Parameter1 of

                0 : begin
                      CSTYLE:=[];            // turn off attributes - normal video
                      FCOLOR:=clBlack;
                      BCOLOR:=clWhite;
                    end;

                1 : begin
                      CSTYLE:=[fsBold];      // highlight
                      FCOLOR:=clBlack;
                      BCOLOR:=clWhite;
                    end;

                4 : begin
                      CSTYLE:=[fsUnderline]; // turn on underline mode
                      FCOLOR:=clBlack;
                      BCOLOR:=clWhite;
                    end;

                5 : begin
                      CSTYLE:=[fsItalic];    // blink
                      FCOLOR:=clBlack;
                      BCOLOR:=clWhite;
                    end;

                7 : begin
                      CSTYLE:=[fsBold];      // turn on inverse video mode
                      FCOLOR:=clBlue;
                      BCOLOR:=clWhite;
                    end;

              else
                ESC_SonderZeichen_in_Matrix('['+IntToStr(CSI_Parameter1)+';'+
                                            IntToStr(CSI_Parameter2)+'m');
              end;
              CSI_Parameter1:=0;
              CSI_Parameter2:=0;
              TermMode:=Normal_TM;
            end; { $6D }

    else
      ESC_SonderZeichen_in_Matrix('['+IntToStr(CSI_Parameter1)+';'+
                                  IntToStr(CSI_Parameter2)+char(Zeichen));
      TermMode:=Normal_TM;
    end;
  end; { procedure IBM_CSI_Par_Zeile_Terminal_Mode }







  // **********************************
  // PC1715 ESC Terminal_Mode
  // **********************************



  procedure PC1715_ESC_Terminal_Mode;
  begin
    if Zeichen>=$80 then begin
      ESC_Zeile:=Zeichen-$80;
      TermMode:=PC1715_TM_Zeile;
    end
    else begin
      ESC_SonderZeichen_in_Matrix('');
      TermMode:=Normal_TM;
    end;
  end; { procedure PC1715_ESC_Terminal_Mode }




  procedure PC1715_ESC_Zeile_Terminal_Mode;
  begin
    if Zeichen>=$80 then begin
      SET_CURSOR_POSITION(ESC_Zeile, Zeichen-$80);
    end
    else begin
      ESC_SonderZeichen_in_Matrix(CHR(ESC_Zeile+$80));
    end;

    TermMode:=Normal_TM;

  end; { procedure PC1715_ESC_Zeile_Terminal_Mode }




  procedure DCA_Terminal_Mode;
  begin
    if Zeichen>=$20 then begin
      ESC_Zeile:=Zeichen-$20;
      TermMode:=DCA_TM_Zeile;
    end
    else begin
      ESC_SonderZeichen_in_Matrix('Y');
      TermMode:=Normal_TM;
    end;
  end; { procedure DCA_Terminal_Mode }





  procedure DCA_Terminal_Mode_Zeile;
  begin
    if Zeichen>=$20 then begin
      SET_CURSOR_POSITION(ESC_Zeile, Zeichen-$20);
    end
    else begin
      ESC_SonderZeichen_in_Matrix('Y'+chr(ESC_Zeile));
    end;
    TermMode:=Normal_TM;
  end; { procedure DCA_Terminal_Mode_Zeile }



begin

  Timer_CURSOR.Enabled:=false;





  if TForm_BUTTONS.CheckBox_LOG_ESC.Checked then begin

    if TermMode=Normal_TM then begin

      if Zeichen<>$1B then begin

        TForm_BUTTONS.Memo_Protokoll.Lines.Add(char(Zeichen));

      end;

    end;

  end;




  if MCursor.on_off then begin
    ZeichenMatrix[MCursor.Zeile,MCursor.Spalte]:=MCursor.Zeichen;
    MCursor.on_off:=false;
  end; { if MCursor.on_off then }


  case TermMode of
      Normal_TM            : Normal_Terminal_Mode;
      IBM_ESC_TM           : IBM_ESC_Terminal_Mode;
      IBM_CSI_TM           : IBM_CSI_Terminal_Mode;
      IBM_CSI_TM_Par       : IBM_CSI_Par_Terminal_Mode;
      IBM_CSI_TM_Par_Zeile : IBM_CSI_Par_Zeile_Terminal_Mode;
      PC1715_ESC_TM        : PC1715_ESC_Terminal_Mode;
      PC1715_TM_Zeile      : PC1715_ESC_Zeile_Terminal_Mode;
      DCA_TM               : DCA_Terminal_Mode;
      DCA_TM_Zeile         : DCA_Terminal_Mode_Zeile;
      ADM31_ESC_TM         : ADM31_ESC_Terminal_Mode;
      ADM31_ESC_TM_R       : ADM31_ESC_Par_Zeile_Terminal_Mode;
      ADM31_ESC_TM_C       : ADM31_ESC_Par_Spalte_Terminal_Mode;
      ADM31_ESC_ESC        : ADM31_ESC_ESC_Terminal_Mode;
      ADM31_ESC_ESC_ESC    : ADM31_ESC_ESC_ESC_Terminal_Mode;
      ADM31_ESC_TM_G       : ADM31_ESC_G_Terminal_Mode;
  end; { case TermMode of }

  if TermMode=Normal_TM then begin
    if TForm_BUTTONS.CheckBox_LOG_ESC.Checked then begin

      if Length(CONOUT_Line)>0 then begin
        CONOUT_Line:=CONOUT_Line+'['+IntToHex(Zeichen,2)+']';
        TForm_BUTTONS.Memo_Protokoll.Lines.Add(CONOUT_Line);
      end;

    end; { if TForm_BUTTONS.CheckBox_LOG_ESC.Checked then }
    CONOUT_Line:='';
  end
  else begin
    CONOUT_Line:=CONOUT_Line+'['+IntToHex(Zeichen,2)+']';
  end; { if TermMode=Normal_TM then }

  Timer_CURSOR.Enabled:=true;

end; { procedure TForm_EMUZ80.Zeichen_an_CONOUT }






procedure TForm_EMUZ80.FormKeyDown(Sender: TObject; var Key: Word;
  Shift: TShiftState);
var
  Zeichen : Byte;
begin

  Zeichen := $00;

  if Shift = [] then begin
    case Key of
      08 : begin // DEL
             Zeichen := $7F;

             RUN_DEL;
           end; { 08 }

      09 : Zeichen := Key; // TAB
      13 : Zeichen := Key; // ENTER

      27 : Zeichen := $1B;  // ESC
      32 : Zeichen := $20; // SPACE

      33 : Zeichen := $12; // Ctrl R
      34 : Zeichen := $03; // Ctrl C

      37 : Zeichen := $13; // links
      38 : Zeichen := $05; // oben
      39 : Zeichen := $04; // rechts
      40 : Zeichen := $18; // unten

      45 : Zeichen := $16; // Einfg = Ctrl V
      46 : Zeichen := $07; // Entf = Ctrl G

      48..57 : Zeichen := Key; // 1..0
      65..90 : Zeichen := Key+32; // a..z
      186 : Zeichen := $75; // ue
      187 : Zeichen := $2B; // +
      188 : Zeichen := $2C; // ,
      189 : Zeichen := $2D; // -
      190 : Zeichen := $2E; // .
      191 : Zeichen := $23; // #
      192 : Zeichen := $6F; // oe
      219 : Zeichen := $73; // s
      220 : Zeichen := $5E; // ^
      221 : Zeichen := $60; // `
      222 : Zeichen := $61; // a
      226 : Zeichen := $3C; // <
    else
      Zeichen := $00;

    end; { case Key of }
  end; { if Shift = [] then }


  if Shift = [SSSHIFT] then begin
    if Key<>16 then begin  // Shift-Taste
      case Key of
        00 : Zeichen := Key;
        48 : Zeichen := $3D; // =
        49 : Zeichen := $21; // !
        50 : Zeichen := $22; // "
        51 : Zeichen := $23; // §
        52 : Zeichen := $24; // $
        53 : Zeichen := $25; // %
        54 : Zeichen := $26; // &
        55 : Zeichen := $2F; // /
        56 : Zeichen := $28; // (
        57 : Zeichen := $29; // )
        65..90 : Zeichen := Key; // A..Z
       187 : Zeichen := $2A; // *
       188 : Zeichen := $3B; // ;
       189 : Zeichen := $5F; // _
       190 : Zeichen := $3A; // :
       191 : Zeichen := $27; // '

       219 : Zeichen := $3F; // ?
       220 : Zeichen := $7E; // ° -> ~
       221 : Zeichen := $60; // `
       226 : Zeichen := $3E; // >
      else
        Zeichen := $00;

      end; { if Key<>16 then }
    end; { if Key<>16 then }
  end; { if Shift=[SSSHIFT] then }


  if Shift = [SSCTRL] then begin
    if Key<>17 then begin
      if (Key>64) and (Key<91) then
         Zeichen := Key-64;


      TForm_BUTTONS.Memo_Protokoll.Lines.Add('Zeichen= '+IntToHex(Zeichen,2)+'H');


    end; { if Key<>17 then }
  end; { if Shift=[SSCTRL] then }


  if Shift = [ssAlt] then begin
    if Key<>18 then begin
      case Key of
        48 : Zeichen := $7D; // }
        55 : Zeichen := $7B; // {
        56 : Zeichen := $5B; // [
        57 : Zeichen := $5D; // ]
        81 : Zeichen := $40; // @
        187 : Zeichen := $7E; // ~
        219 : Zeichen := $5C; // \
        226 : Zeichen := $7D; // |

      else
        Zeichen := $00;

      end; { case Key of }
    end; { if Key<>17 then }
  end; { if Shift=[SSCTRL] then }




  if Zeichen>$00 then begin
    mod_pio.CONIN_Buffer := mod_pio.CONIN_Buffer + Char(Zeichen);

    Caption := mod_pio.CONIN_Buffer;

  end;


end; { procedure TForm_EMUZ80.FormKeyDown }








{ Console Console Console Console  }








procedure TForm_EMUZ80.Image_Button_EinstellungenClick(Sender: TObject);
begin

  if Form_Button_aktiv then begin
    TForm_BUTTONS.Visible:=false;
    Form_Button_aktiv:=false;
  end
  else begin
    TForm_BUTTONS.Visible:=true;
    Form_Button_aktiv:=true;
  end;

end; { procedure TForm_EMUZ80.Image_Button_EinstellungenClick }








procedure TForm_EMUZ80.Image_Button_SchliessenClick(Sender: TObject);
begin
  Form_EMUZ80.Close;
end; { procedure TForm_EMUZ80.Image_Button_SchliessenClick }





procedure TForm_EMUZ80.FormDestroy(Sender: TObject);
begin

  Timer_CURSOR.Enabled:=false;


  if Assigned(FIni) then begin

    FIni.WriteInteger('FORM','Main_Left', Form_EMUZ80.Left );
    FIni.WriteInteger('FORM','Main_Top', Form_EMUZ80.Top );

    FIni.WriteInteger('FORM','Button_Left', Button_Left );
    FIni.WriteInteger('FORM','Button_Top', Button_Top );

    FIni.WriteInteger('FORM','Mem_Left', Mem_Left );
    FIni.WriteInteger('FORM','Mem_Top', Mem_Top );
    FIni.WriteInteger('FORM','CPU_Left', CPU_Left );
    FIni.WriteInteger('FORM','CPU_Top', CPU_Top );
    FIni.WriteInteger('FORM','TRACE_Left', TRACE_Left );
    FIni.WriteInteger('FORM','TRACE_Top', TRACE_Top );
    FIni.WriteInteger('FORM','TRACE_Width', TRACE_Width );
    FIni.WriteInteger('FORM','TRACE_Height', TRACE_Height );
    FIni.WriteBool('FORM','CPM_AutoStart', CPM_AutoStart );
    FIni.WriteInteger('Emulation','Emu_Mode', Emu_Mode );

    FIni.WriteString('CPM','CPM_DIR_LW_A', CPM_DIR_LW_A );
    FIni.WriteString('CPM','CPM_DIR_LW_B', CPM_DIR_LW_B );

    FIni.UpdateFile;
    FIni.Free;
  end; { if Assigned(FIni) then }




  wcpm_LW_A.Export_Files_From_Image;
  wcpm_LW_A.Destroy;

  wcpm_LW_B.Export_Files_From_Image;
  wcpm_LW_B.Destroy;

  Z80.Free;

  mod_pio.Free;


end; { procedure TForm_EMUZ80.FormDestroy }






initialization
  {$I MainForm_EMUZ80.lrs}

end.

